// sockutil.cc
// code for sockutil.h
// copyright SafeTP Development Group, Inc., 2000  Terms of use are as specified in license.txt

#include "xassert.h"      // xassert
#include <string.h>       // strlen, memset
#include <stdio.h>        // printf
#include <stdlib.h>       // exit, rand, srand

#include "socket.h"       // socket functions
#include "strtokp.h"      // StrtokParse
#include "nonport.h"      // portableSleep, getProcessId
#include "sockutil.h"     // this module


// ---------- diagnostic facilities ------------------
// general diagnostic
#ifdef SOCKUTIL_DIAGNOSTICS
#  define DIAGNOSTIC 1
#else
#  define DIAGNOSTIC 0
#endif

// connection activity: socket open, close
#define CONN_DIAGNOSTIC 1

// data activity: send and receive
#define DATA_DIAGNOSTIC 0

// select block/unblock
#define SELECT_DIAGNOSTIC 0

#if DIAGNOSTIC
#  include <iostream.h>      // cout
#  define diagnostic(expr) cout << "socket(" << numSocketsOpen() << "): " << expr << endl
#else
#  define diagnostic(expr) ((void)0)
#endif

#if CONN_DIAGNOSTIC
#  define connDiagnostic(expr) diagnostic(expr)
#else
#  define connDiagnostic(expr) ((void)0)
#endif

#if DATA_DIAGNOSTIC
#  define dataDiagnostic(expr) diagnostic(expr)
#else
#  define dataDiagnostic(expr)
#endif

#if SELECT_DIAGNOSTIC
#  define selectDiagnostic(expr) diagnostic(expr)
#else
#  define selectDiagnostic(expr)
#endif


// ------------------- xSocket ----------------------------
xSocket::xSocket(SOCKET s, char const *msg)
  : xBase(msg),
    socket(s),
    graceful(false),
    temporary(false)
{}

xSocket::xSocket(SOCKET s)
  : xBase("closed gracefully"),
    socket(s),
    graceful(true),
    temporary(false)
{}

xSocket::xSocket(xSocket const &obj)
  : xBase(obj),
    socket(obj.socket),
    graceful(obj.graceful),
    temporary(obj.temporary)
{}

xSocket::~xSocket()
{}


// -------------------- xResolveFailure -----------------------
xResolveFailure::xResolveFailure(char const *hn)
  : xBase(stringb("Failed to resolve hostname \"" << hn << "\"")),
    hostname(hn)
{}

xResolveFailure::xResolveFailure(xResolveFailure const &obj)
  : xBase(obj),
    hostname(obj.hostname)
{}

xResolveFailure::~xResolveFailure()
{}


// ----------------- open-sockets counting ------------------
// for thread safety, reduce visibility by making static and adding
// an underscopre to change the name
static int _socketsOpen = 0;

int numSocketsOpen()
{
  LOCKER
  return _socketsOpen;
}

void incOpenSockets()
{
  LOCKER
  _socketsOpen++;
}

void decOpenSockets()
{
  LOCKER
  _socketsOpen--;
}


// ----------------- other standalone routines ------------------------
char recvChar(SOCKET s)
{
  char c;
  int len = basicRecv(s, &c, 1);
  if (len == 0) {
    // graceful close
    xsocketClosed(s);
  }

  return c;
}


// not efficient at all..
string recvLine(SOCKET s)
{
  // phasing out usage, so want to see usage
  //breaker();
  //printf(" ** recvLine called **\n");

  // (reasonably) efficient growing string
  stringBuilder buf;

  // read 2 chars, since we can't begin testing
  // for a final CRLF until there are at least 2
  buf << recvChar(s);
  buf << recvChar(s);
  int len = 2;

  // test..
  while (!( buf[len-2]=='\r' && buf[len-1]=='\n' )) {
    // read another character
    buf << recvChar(s);
    len++;
  }

  return string(buf, len-2);
    // return string without the EOL sequence
}


// NOTE: the network buffers are only guaranteed to hold 1 byte for
// peeking.  virtually all implementations can hold more, typically 1k
// or 4k.  but that's still not infinite, so this should only be used
// in situations where the expected string is short.
string peekLine(SOCKET s)
{
  enum { BUFSIZE=1024 };    // arbitrary maximum (but remember we expect short strings)
  char buf[BUFSIZE];

  char const *crlf;
  do {
    // grab what's in the buffer
    int len = recv(s, buf, BUFSIZE-1, MSG_PEEK);
    if (len == -1) {
      xsocket(s, "recv");
    }
    buf[len] = 0;   // null terminate so can use strstr

    // see if a CRLF is in the buffer
    crlf = strstr(buf, "\r\n");
    if (!crlf) {
      // it's not there

      // if what we read maxed-out the buffer
      if (len == BUFSIZE-1) {            
        // then we're not going to find the CRLF.. I don't know
        // a good action to take here
        xfailure("peekLine: the buffer filled without seeing a complete line");
      }

      // wait a little bit before checking again
      portableSleep(1);
    }
  } while (!crlf);

  // ok, a string was found; return it (without CRLF)
  return string(buf, crlf-buf);
}


unsigned long recvNBO32(SOCKET s)
{
  unsigned char buf[4];
  recvAll(s, (char*)buf, 4);

  // construct from big-endian
  return (buf[0] << 24) |
         (buf[1] << 16) |
         (buf[2] << 8) |
         buf[3];
}


void sendNBO32(SOCKET s, unsigned long value)
{
  // represent as big-endian
  byte buf[4];
  buf[0] = (byte)((value >> 24) & 0xff);
  buf[1] = (byte)((value >> 16) & 0xff);
  buf[2] = (byte)((value >> 8) & 0xff);
  buf[3] = (byte)(value & 0xff);

  //printf("NBO32 sending: %X %X %X %X\n",
  //       buf[0], buf[1], buf[2], buf[3]);

  sendAll(s, (char const*)buf, 4);
}


void recvAll(SOCKET s, char *buf, int len)
{
  int result = recvAllToEOF(s, buf, len);
  if (result < len) {
    // this fn throws an exception on EOF
    xsocketClosed(s);
  }
}


int basicRecv(SOCKET s, char *buf, int len)
{
  // block until at least 1 byte read
  xassert(buf != NULL && len >= 1);
  int received = recv(s, buf, len, 0);
  if (received == SOCKET_ERROR) {
    xsocket(s, "recv");
  }
  return received;
}


int recvAllToEOF(SOCKET s, char *buf, int len)
{
  int totalRead = 0;
  while (len != 0) {
    // block until at least 1 byte read
    int xmitted = basicRecv(s, buf, len);
    if (xmitted == 0) {
      // EOF
      break;
    }

    // quick check to guard against recv writing past end
    // due to math error in here
    xassert(xmitted <= len);

    // advance counters
    buf += xmitted;
    len -= xmitted;
    totalRead += xmitted;
  }

  return totalRead;
}


void sendChar(SOCKET s, char c)
{
  int len = basicSend(s, &c, 1);
  if (len == 0) {
    // graceful closure
    xsocketClosed(s);
  }
}


// send all of the specified buffer, blocking
// as necessary
void sendAll(SOCKET s, char const *buf, int len)
{
  while (len != 0) {
    // is send guaranteed to block until at least 1 byte sent?  yes
    int sent = basicSend(s, buf, len);
    if (sent == 0) {
      xsocketClosed(s);
    }
    xassert(sent <= len);
      // can't send more than was in buffer...

    buf += sent;
    len -= sent;
  }
}


void sendAllString(SOCKET s, char const *str)
{
  sendAll(s, str, strlen(str));
}


void sendEOL(SOCKET s)
{
  static char const eol[] = "\r\n";
  sendAll(s, eol, 2);
}


int basicSend(SOCKET s, char const *data, int len)
{
  xassert(data != NULL && len >= 1);
  int sent = send(s, data, len, 0);
  if (sent == SOCKET_ERROR) {
    xsocket(s, "send");
  }
  return sent;
}


// create a socket, and possibly set SO_RESUSEADDR
SOCKET create_socket()
{
  // create the socket
  SOCKET s = socket(AF_INET, SOCK_STREAM, 0);
  if (s == INVALID_SOCKET) {
    xsocket(s, "socket");
  }
  incOpenSockets();

  // during development, it's annoying when the harder crashes cause
  // us to not be able to reuse a port for a while; it also increases
  // the # of possible connections for bind-range-of-ports servers
  // (I'm questioning this decision, since it affects TCP's reliability,
  // but I can't think of a good way to handle it..)
  //
  // it turns out that on win32 SO_REUSEADDR's semantics are different,
  // namely that it *does* allow two sockets to listen to the same
  // address (unlike all unix implementations I've encountered); since
  // this can especially be a problem with the -r switch, we'll skip
  // this on win32; I also added another escape hatch to disable if
  // necessary elsewhere
  #if !defined(__WIN32__) && !defined(DONT_REUSEADDR)
  {
    int boolval = 1;
    int res = setsockopt(
      s /*socket*/,
      SOL_SOCKET /*level*/,
      SO_REUSEADDR /*optname*/,    // reuse if possible
      (char*)&boolval /*optval*/,
      sizeof(boolval) /*optlen*/);

    if (res == SOCKET_ERROR) {
      xsocket(s, "setsockopt", true /*close*/);
    }
  }
  #endif

  return s;
}


// return false on failure
bool inner_bind_socket(SOCKET s, IPAddress interface, int port)
{
  // create the address structure
  sockaddr_in saddr;
  memset((char *)&saddr, 0, sizeof(saddr));
  saddr.sin_family     	= AF_INET;
  saddr.sin_addr.s_addr = htonl(interface);      // which interface to bind to
  saddr.sin_port       	= htons((short)port);    // correct byte order

  // attach that local address to the socket
  return bind(s, (sockaddr *)&saddr, sizeof(saddr)) != SOCKET_ERROR;
}


// throw exception on failure
void bind_socket(SOCKET s, IPAddress interface, int port)
{
  if (!inner_bind_socket(s, interface, port)) {
    xsocket(s, "bind", true /*close*/);
  }
}


void bind_socket_range(SOCKET s, IPAddress interface, PortRange const &range)
{
  if (!range.restricted) {
    // binding to PORT_ANY makes it unrestricted
    bind_socket(s, interface, PORT_ANY);
    return;
  }

  // pick a port to try; use randomness instead of maintaining state
  int maxTries = range.high - range.low + 1;
  int port = range.low + rand() % maxTries;

  // we will as many attempts as there are ports in the range
  for (int attempt = 0; attempt < maxTries; attempt++) {
    // increment to next port
    // (originally I was picking a random port for each attempt; Dan
    // suggests that when most ports are used sequential scan may be
    // better)
    port++;
    if (port > range.high) {
      port = range.low;
    }
           
    if (inner_bind_socket(s, interface, port)) {
      return;
    }
  }

  // give up
  xSocket x(s, stringb("Failed to bind any ports in the range [" <<
                       range.low << "," << range.high << "] after " <<
                       maxTries << " attempts; giving up."));
  x.temporary = true;    // hopefully
  THROW(x);
}

 
// put a socket into a listening state; 'interface' is passed
// so we can print some diagnostics
void listen_socket(SOCKET s, IPAddress interface)
{
  if (listen(s, 5 /*backlog queue size*/) == SOCKET_ERROR) {
    xsocket(s, "listen", true /*close*/);
  }

  if (interface == INADDR_ANY) {
    // can't just print 'port' because it could be PORT_ANY
    connDiagnostic("listening on port " << getLocalPort(s) <<
                   " on all interfaces");
  }
  else {
    connDiagnostic("listening on port " << getLocalPort(s) <<
                   " on interface " << formatAddress(interface));
  }
}


// create a listener socket on the specified port and interface
SOCKET interface_listen_socket(IPAddress interface, int port)
{
  // create the socket and bind it to the interface and port
  SOCKET s = create_socket();
  bind_socket(s, interface, port);

  // establish that this socket is listening for incoming connections
  listen_socket(s, interface);

  return s;
}


SOCKET interface_listen_socket_range(IPAddress interface,
                                     PortRange const &range)
{
  // same sequence of calls as interface_listen_socket
  SOCKET s = create_socket();
  bind_socket_range(s, interface, range);
  listen_socket(s, interface);
  return s;
}


#ifndef SOCKLEN_TYPE                 
  // for linux, set it to socklen_t; for other systems, sometimes
  // it's unsigned long, sometimes it's int; we'll just guess
  // socklen_t if nothing is specified
  #define SOCKLEN_TYPE socklen_t
#endif


// accept a connection
SOCKET accept_socket(SOCKET s)
{
  sockaddr saddr;
  SOCKLEN_TYPE saddrLen = sizeof(saddr);
  SOCKET ret = accept(s, &saddr, &saddrLen);

  if (ret == INVALID_SOCKET) {
    xsocket(s, "accept");
  }
  incOpenSockets();

  connDiagnostic("accepted connection: " << sockInfo(ret));

  return ret;
}


// format an IP address into normal decimal-dot notation
string formatAddress(IPAddress addr)
{
  return stringb( ((addr >> 24) & 0xff) << "." <<
                  ((addr >> 16) & 0xff) << "." <<
		  ((addr >> 8) & 0xff) << "." <<
		  (addr & 0xff) );
}


string formatAddress(IPAddress addr, int port)
{
  return stringb(
    formatAddress(addr) << ", port " << port );
}


// format a socket address into a string
string formatAddress(sockaddr_in const &saddr)
{
  // extract fields
  IPAddress ip = ntohl(saddr.sin_addr.s_addr);
  int port = ntohs(saddr.sin_port);

  // write buffer
  return formatAddress(ip, port);
}


// initiate a connection
SOCKET connect_socket(IPAddress ipaddr, int port, PortRange const *range)
{
  // create address structure
  sockaddr_in saddr;      // address of receiver
  memset((char *)&saddr, 0, sizeof(saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_port   = htons((short)port);    // get correct byte ordering
  saddr.sin_addr.s_addr = htonl(ipaddr);

  // create the socket
  SOCKET s = create_socket();

  // bind the socket locally if we've been asked to do so by 'range'
  if (range) {
    bind_socket_range(s, ipaddr, *range);
  }

  // connect the socket to the resolved address and port (sockaddr)
  if (connect(s, (sockaddr *)&saddr, sizeof(saddr))
       == SOCKET_ERROR) {
    xsocket(s, "connect", true /*close*/);
  }

  connDiagnostic("connected: " << sockInfo(s));

  return s;
}


void close_socket(SOCKET s)
{
  #if DIAGNOSTIC && CONN_DIAGNOSTIC
    int port;
    try {
      port = getLocalPort(s);
      connDiagnostic("closing: " << sockInfo(s));
    }
    catch (...)
      {}    // I don't care--the next will probably throw anyway
  #endif

  if (closesocket(s) == SOCKET_ERROR) {
    xsocket(s, "closesocket");
  }
  decOpenSockets();
}


bool pollReadable(SOCKET s)
{
  // make a SocketSet of one, and use its fn
  SocketSet set;
  set.add(s);
  return set.pollReadable() > 0;   // it returns # of readable sockets
}


bool isClosed(SOCKET s)
{
  // first, test to see if recv will block
  if (!pollReadable(s)) {
    // isn't readable, therefore can't be closed
    // (or at least our OS can't have detected it)
    return false;
  }

  char c;
  int len = recv(s, &c, 1, MSG_PEEK);
  if (len == 0) {
    // socket closed
    connDiagnostic("detected closed: " << sockInfo(s));
    return true;
  }
  else if (len == SOCKET_ERROR) {
    // some error
    xsocket(s, "recv(MSG_PEEK)");
  }

  // not closed
  return false;
}


// throw a particular exception if it's closed
void checkClosed(SOCKET s)
{
  if (isClosed(s)) {
    xsocketClosed(s);
  }
}


void getSockName(SOCKET s, sockaddr_in &addr)
{
  SOCKLEN_TYPE namelen = sizeof(sockaddr_in);    
  if (getsockname(s, (sockaddr*)&addr, &namelen) == SOCKET_ERROR) {
    xsocket(s, "getsockname");
  }
}

IPAddress getLocalAddress(SOCKET s)
{
  sockaddr_in saddr;
  getSockName(s, saddr);
  return ntohl(saddr.sin_addr.s_addr);
}

int getLocalPort(SOCKET s)
{
  sockaddr_in saddr;
  getSockName(s, saddr);
  return ntohs(saddr.sin_port);
}


void getSockPeer(SOCKET s, sockaddr_in &addr)
{
  SOCKLEN_TYPE namelen = sizeof(sockaddr_in);     
  if (getpeername(s, (sockaddr*)&addr, &namelen) == SOCKET_ERROR) {
    xsocket(s, "getpeername");
  }
}

IPAddress getRemoteAddress(SOCKET s)
{
  sockaddr_in saddr;
  getSockPeer(s, saddr);
  return ntohl(saddr.sin_addr.s_addr);
}

int getRemotePort(SOCKET s)
{
  sockaddr_in saddr;
  getSockPeer(s, saddr);
  return ntohs(saddr.sin_port);
}


void printSocketInfo(SOCKET s)
{
  sockaddr_in saddr;

  // get, display local name
  getSockName(s, saddr);
  printf("  local name: %s\n", (char const*)formatAddress(saddr));

  // get, display remote name
  try {
    getSockPeer(s, saddr);
    printf("  remote name: %s\n", (char const*)formatAddress(saddr));
  }
  catch (...) {
    printf("  (not connected)\n");
  }
}


bool isListening(SOCKET s)
{
  #ifdef SO_ACCEPTCONN
    long retval;
    SOCKLEN_TYPE retvalLen = sizeof(retval);
    int err = getsockopt(s, SOL_SOCKET, SO_ACCEPTCONN,
                         (char*)&retval, &retvalLen);
    if (err == SOCKET_ERROR) {
      xsocket(s, "getsockopt");
    }

    return retval != 0;

  #else
    // turns out some systems don't have an easy way to detect
    // if something is a listener (maybe winsock is only that does?)
    // for now, I can tolerate simply not having this information
    return false;
  #endif
}


// this is a separate fn so we can use 'return' to
// end collection of information but finish sockInfo normally
void sockInfoHelper(SOCKET s, stringBuilder &sb)
{
  // test for special value
  if (s == INVALID_SOCKET) {
    sb << "invalid socket";
    return;
  }

  // local address
  sb << formatAddress(getLocalAddress(s)) <<
        ":" << getLocalPort(s) << " - ";

  // listening?
  if (isListening(s)) {
    sb << "(listener)";
    return;
  }

  // remote address
  sb << formatAddress(getRemoteAddress(s)) <<
        ":" << getRemotePort(s);
}


// catch-all info retrieval
string sockInfo(SOCKET s)
{
  // for accumulating results
  stringBuilder sb;
  sb << "[";

  // I don't want exceptions here dirtying up my output
  bool prev = xBase::logExceptions;
  xBase::logExceptions = false;

  try {
    sockInfoHelper(s, sb);
  }

  catch (xBase &x) {
    sb << "(" << x.why() << ")";
  }

  xBase::logExceptions = prev;

  sb << "]";
  return sb;
}


int getServByName(char const *serviceName)
{
  struct servent *se = getservbyname(serviceName, "tcp");
  if (!se) {
    return 0;    // not found
  }

  int port = ntohs((short)se->s_port);
  return port;
}


IPAddress resolveHostName(char const *hostname)
{
  xassert(hostname);

  // check for dotted decimal
  try {                               
    // if it is dotted decimal, this will simply parse it and return
    return parseDottedDecimal(hostname);
  }
  catch (...)
    {}       // not a valid decimal addr; go on to try DNS

  // invoke the Domain Name Service (DNS)
  // NOTE: This is a blocking operation, and may timeout only after
  //       a substantial period of time.
  hostent *he = gethostbyname(hostname);

  if (he == NULL) {
    xResolveFailure x(hostname);
    THROW(x);
  }
  else {
    // if multiple addresses are returned, we only return the first
    return ntohl( *((unsigned long const*)he->h_addr_list[0]) );
  }
}


// --------------------- parsing ----------------------
IPAddress parseDottedDecimal(char const *decimalAddr)
{
  IPAddress ret = ntohl(inet_addr(decimalAddr));
  if (ret == INADDR_NONE) {
    xfailure("inet_addr: invalid dotted-decimal address");
  }

  return ret;


  #if 0
  // this is supposedly the right way but some systems don't
  // yet have inet_aton
  struct in_addr addr;
  if (!inet_aton(decimalAddr, &addr)) {
    xfailure("inet_aton: invalid dotted-decimal address");
  }

  return ntohl(addr.s_addr);
  #endif // 0
}


void parseAddrAndPort(IPAddress &ipaddr, int &port, char const *addrAndPort)
{
  // separate on a colon, if present
  StrtokParse tok(addrAndPort, ":");

  if (tok == 2) {
    // has both an IP address and a port
    ipaddr = parseDottedDecimal(tok[0]);
    port = atoi(tok[1]);
  }

  else if (tok == 1) {
    // only a port is supplied
    ipaddr = INADDR_ANY;
    port = atoi(tok[0]);
  }

  else {
    // wrong format
    xfailure("parseAddrAndPort: address must be AAA.BBB.CCC.DDD:PPPP or PPPP format");
  }
}


// ------------------ SocketSet ------------------------------------
// timeout of 0 means poll without blocking
// (this is expected to be const, but I don't tag
// it as such for same reason 'timeout' parameter
// to select is not so tagged (see sockutil.h))
timeval SocketSet::timeZero = {0, 0};

SocketSet::SocketSet()
{
  reset();
}

SocketSet::~SocketSet()
{}


bool SocketSet::contains(SOCKET s) const
{
  // on unix, we segfault if we test INVALID_SOCKET
  if (s == INVALID_SOCKET) {
    return false;
  }
                                          
  // FD_ISSET declared incorrectly in OSF1 system header files (doesn't mention const for fd_set*)
  return !!FD_ISSET(s, ((fd_set*)&set));
}


void SocketSet::add(SOCKET s)
{
  // this fn won't work after select is called
  xassert(!used);

  // silently filter out INVALID_SOCKETs
  if (s == INVALID_SOCKET) {
    return;
  }

  FD_SET(s, &set);
  maxSocket = mymax(maxSocket, (unsigned)s);
  numSockets++;
}


void SocketSet::reset()
{
  FD_ZERO(&set);
  maxSocket = 0;
  used = false;
  numSockets = 0;
}


int SocketSet::select(bool read, bool write, bool except,
                      /*const*/ struct timeval *timeout)
{
  xassert(!used);
  xassert(numSockets > 0);   // need at least one valid socket!

  selectDiagnostic("blocked on select()");
  int result = ::select(
    maxSocket+1,                      // nfds
    read? &set : (fd_set*)NULL,       // readfds
    write? &set : (fd_set*)NULL,      // writefds
    except? &set : (fd_set*)NULL,     // exceptfds
    timeout);
  if (result == SOCKET_ERROR) {
    xsocket((SOCKET)NULL, "select");
  }
  selectDiagnostic("select() unblocked");

  used = true;
  return result;
}


// ----------------- xsocket -----------------------
// compose a complete error message string
string socketErrString(char const *msg)
{
  int code = getSocketErrorCode();
  char const *codetext = errorCodeString(code);
  if (!codetext) {
    codetext = "unknown error";
  }

  return stringb(
    msg << ": " <<
    codetext <<
    " (code " << code << ")");
}


// handle an error
// (originally had 3rd param named 'close'; that is a problem
// on unix, where 'closesocket' is #defined to 'close')
void xsocket(SOCKET s, char const *msg, bool doClose)
{
  xSocket x(s, socketErrString(msg));

  if (doClose &&
      closesocket(s) != SOCKET_ERROR) {
    decOpenSockets();
  }

  THROW(x);
}


void xsocketClosed(SOCKET s)
{
  xSocket tmp(s);
  THROW(tmp);
}


// ------------------- endian junk ---------------------
// force the definition of either SAFETP_LITTLE_ENDIAN or SAFETP_BIG_ENDIAN
#if !defined(SAFETP_LITTLE_ENDIAN) && !defined(SAFETP_BIG_ENDIAN)
#  error define SAFETP_LITTLE_ENDIAN or SAFETP_BIG_ENDIAN
#endif

// return true if this is a big-endian machine, false if is
// little-endian (reference to endianness discussion might
// be nice here..)
#if 0   // this code crashes gcc-2.95.2 -O2 on solaris/sparc
bool isBigEndian()
{
  union {
    int i;                  // machine word
    byte b[sizeof(int)];    // b[0] overlaid with first byte of i
  };
  i = 0xFF;    // set lsb, zero all others
  return b[0] != 0xFF;
}

#else   // 1
// x must be global to work around an egcs-1.1.2 optimizer bug
static int endianTestValue = 0;     // start with all bits 0
bool isBigEndian()
{
  // little endian: sets endianTestValue to 1
  // big endian: sets endianTestValue to a large number (e.g. 0x01000000 if 32 bits)
  *((char*)(&endianTestValue)) = 1;

  if (endianTestValue == 1) {
    return false;   // little
  }
  else {
    return true;    // big
  }
}
#endif  // 0/1


// --------------- nonportable routines ----------------------------
#ifdef __WIN32__
// (exported)
// return string associated with a given code
char const *errorCodeString(int code)
{
  static struct {
    int code;
    char const *message;
  } const arr[] = {
    { WSAEACCES,          "Broadcast address requested, proper flags not set" },
    { WSAEADDRINUSE,      "Specified address already in use" },
    { WSAEADDRNOTAVAIL,   "Address not available or accessible" },
    { WSAEAFNOSUPPORT,    "Address format not supported" },
    { WSAEALREADY,        "\"already\" something (?)" },    // already what?
    { WSAEBADF,           "Bad file number" },
    { WSAECONNABORTED,    "Connection aborted, possibly due to timeout" },
    { WSAECONNREFUSED,    "Connection refused" },
    { WSAECONNRESET,      "Connection reset by remote side" },
    { WSAEDESTADDRREQ,    "Destination address required" },
    { WSAEDISCON,         "Socket disconnected" },
    { WSAEDQUOT,          "\"DQUOT\" (?)" },     // what does this mean?
    { WSAEFAULT,          "Memory buffer too small" },
    { WSAEHOSTDOWN,       "Remote host is down" },
    { WSAEHOSTUNREACH,    "Remote host is unreachable" },
    { WSAEINPROGRESS,     "Blocking call in progress" },
    { WSAEINTR,           "Blocking call interrupted" },
    { WSAEINVAL,          "Invalid parameter" },
    { WSAEISCONN,         "Socket already connected" },
    { WSAELOOP,           "Routing loop discovered (?)" },
    { WSAEMFILE,          "No more file descriptors available" },
    { WSAEMSGSIZE,        "Message size exceeds maximum datagram size" },
    { WSAENAMETOOLONG,    "Name too long" },
    { WSAENETDOWN,        "Network subsystem failure detected" },
    { WSAENETRESET,       "Connection was reset" },
    { WSAENETUNREACH,     "Network is unreachable" },
    { WSAENOBUFS,         "Insufficient buffer space available" },
    { WSAENOPROTOOPT,     "Invalid protocol option" },
    { WSAENOTCONN,        "Socket is not connected" },
    { WSAENOTEMPTY,       "Buffer not empty" },
    { WSAENOTSOCK,        "Connection is already closed" },
    { WSAEOPNOTSUPP,      "Operation not supported (e.g. OOB)" },
    { WSAEPFNOSUPPORT,    "Protocol family not supported" },
    { WSAEPROCLIM,        "\"PROCLIM\" (?)" },    // anyone know what this means?
    { WSAEPROTONOSUPPORT, "The specified protocol is not supported" },
    { WSAEPROTOTYPE,      "Specified protocol is wrong type" },
    { WSAEREMOTE,         "Error at remote host" },
    { WSAESHUTDOWN,       "Socket has been shut down" },
    { WSAESOCKTNOSUPPORT, "Socket type not supported" },
    { WSAESTALE,          "Stale handle" },
    { WSAETIMEDOUT,       "Connection timed out" },
    { WSAETOOMANYREFS,    "Too many references" },
    { WSAEUSERS,          "Too many users" },
    { WSAEWOULDBLOCK,     "Asynchronous operation cannot be scheduled" },
    { WSAHOST_NOT_FOUND,  "Authoritative Answer Host not found" },
    { WSANO_ADDRESS,      "No address, look for MX record (?)" },
    { WSANO_DATA,         "Valid name, no data record of requested type" },
    { WSANO_RECOVERY,     "Non-recoverable error" },
    { WSANOTINITIALISED,  "Winsock not initialized" },
    { WSASYSNOTREADY,     "Network subsystem not ready" },
    { WSATRY_AGAIN,       "Non-Authoritative Host not found" },
    { WSAVERNOTSUPPORTED, "Winsock version not supported" },

    { 0,                  "No error" }
  };

  loopi(TABLESIZE(arr)) {
    if (arr[i].code == code) {
      return arr[i].message;
    }
  }

  // unknown code
  return NULL;
}


// (exported)
int getSocketErrorCode()
{ 
  return WSAGetLastError();
}


// (exported)
void socket_lib_init()
{
  WSADATA d;
  if (0 != WSAStartup(MAKEWORD(1,1), &d)) {
    xsocket(INVALID_SOCKET, "WSAStartup");
  }

  // randomize
  srand(getProcessId());
}

// (exported)
void socket_lib_deinit()
{
  // added this because Dan says he wants sftpd to do this
  WSACleanup();
}

#else                   //  --  win32 above, unix below  --

#include <errno.h>      // errno, strerror

char const *errorCodeString(int code)
{
  return strerror(code);
}


int getSocketErrorCode()
{
  return errno;
}


void socket_lib_init()
{
  // nothing special required for berkeley sockets

  // but I want to randomize rand() so that two different
  // server processes won't request the same sequence of ports
  // (which would exacerbate the reliability problem with
  // SO_REUSEADDR)
  srand(getProcessId());
}

void socket_lib_deinit()
{
  // nothing special required for berkeley sockets
}

#endif //    ^^ unix


// --------------------- test code ----------------------
#ifdef TEST_SOCKUTIL

#include "test.h"    // USUAL_MAIN

#ifdef __UNIX__

#include <sys/wait.h>     // wait

// actually try to make some connections and send some data
void communicationTests()
{
  printf("start of communicationTests\n");

  int childPid = fork();
  if (childPid != 0) {
    // --------------- parent ---------------
    // don't let this take too long
    alarm(20);

    // listen for a connection from the child
    printf("parent: waiting for child to connect\n");
    SOCKET listener = interface_listen_socket(INADDR_LOOPBACK, 2345);
    SOCKET s = accept_socket(listener);
    close_socket(listener);

    // connected; send something trivial
    printf("parent: sending int\n");
    sendNBO32(s, 0x4567890A);

    // receive something trivial
    printf("parent: waiting for char\n");
    char ch = recvChar(s);
    printf("parent: received char: %c\n", ch);
    xassert(ch == 'x');

    // listen on a range
    printf("parent: making a range listener\n");
    PortRange range(5000,5050);
    listener = interface_listen_socket_range(INADDR_LOOPBACK, range);

    // check that we got a port in the range requested
    int port = getLocalPort(listener);
    printf("parent: got a listener on port %d\n", port);
    xassert(5000 <= port && port <= 5050);

    // tell the client which port we got
    sendNBO32(s, (unsigned long)port);
    
    // wait for the connection
    SOCKET t = accept_socket(listener);
    close_socket(listener);
    printf("parent: got child's second connection\n");
    
    // close everything down
    close_socket(t);
    close_socket(s);

    // wait for child to exit
    int status;
    wait(&status);
    printf("parent: child exited with code %d\n", status);
    xassert(status == 0);
  }
  else {
    // ---------------- child ---------------
    // wait for parent to be ready to accept a connection
    sleep(1);

    // connect to parent
    printf("child: connecting\n");
    SOCKET s = connect_socket(INADDR_LOOPBACK, 2345);

    // receive the integer
    printf("child: connected; waiting for int\n");
    unsigned long val = recvNBO32(s);
    printf("child: received 0x%lX\n", val);
    xassert(val == 0x4567890A);

    // send a char
    printf("child: sending char\n");
    sendChar(s, 'x');

    // receive port that parent is now listening at
    int remotePort = (int)recvNBO32(s);

    // connect to it, using a range for my local port
    printf("child: connecting to parent at port %d\n", remotePort);
    PortRange range(6000,6070);
    SOCKET t = connect_socket(INADDR_LOOPBACK, remotePort, &range);

    // verify range
    int localPort = getLocalPort(t);
    printf("child: my local port is %d\n", localPort);
    xassert(6000 <= localPort && localPort <= 6070);

    // close it all down
    close_socket(t);
    close_socket(s);

    exit(0);
  }
  
  printf("end of communicationTests\n");
}    

#else  // windows
void communicationTests()
{
  // windows doesn't have fork (right?) so this would be more painful..
  // forget it
}
#endif // !__UNIX__


void testParseDottedDecimal(char const *decimal, IPAddress expected)
{
  try {
    IPAddress got = parseDottedDecimal(decimal);
    xassert(got == expected);
  }
  catch (...) {
    // did we expect the failure?
    xassert(expected == INADDR_NONE);
  }
}


void testParseAddrAndPort(char const *addrAndPort,
                          IPAddress expAddr, int expPort)
{
  try {
    IPAddress gotAddr;
    int gotPort;
    parseAddrAndPort(gotAddr, gotPort, addrAndPort);
    xassert(gotAddr == expAddr && gotPort == expPort);
  }
  catch (...) {
    xassert(expAddr == INADDR_NONE);
  }
}


int entry()
{
  socket_lib_init();

  // reduce clutter
  xBase::logExceptions = false;

  // verify endianness
  #if defined(SAFETP_LITTLE_ENDIAN)
    if (isBigEndian()) {
      xfailure("SAFETP_LITTLE_ENDIAN is defined but isBigEndian() returns true");
    }
  #elif defined(SAFETP_BIG_ENDIAN)
    if (!isBigEndian()) {
      xfailure("SAFETP_BIG_ENDIAN is defined but isBigEndian() returns false");
    }
  #else
    xfailure("neither SAFETP_LITTLE_ENDIAN nor SAFETP_BIG_ENDIAN are defined");
  #endif

  // test vectors for parseDottedDecimal
  testParseDottedDecimal("1.2.3.4", 0x01020304);
  testParseDottedDecimal("0.0.0.0", 0);
  testParseDottedDecimal("255.255.255.255", 0xffffffff);
  testParseDottedDecimal("1234", INADDR_NONE);

  // test vectors for parseAddrAndPort
  testParseAddrAndPort("1.2.3.4:5", 0x01020304, 5);
  testParseAddrAndPort("67", INADDR_ANY, 67);
  testParseAddrAndPort(":", INADDR_NONE, 0);
  testParseAddrAndPort("2.3.4.5:", INADDR_NONE, 0);

  // verify that binding the same port twice doesn't work (at least
  // from the same process...)
  SOCKET s = listen_socket(3456);
  try {
    listen_socket(3456);
    xfailure("it let me bind the same port twice!");
  }
  catch (xSocket&) {
    printf("good: it didn't let me bind the same port twice\n");
  }
  close_socket(s);

  communicationTests();

  printf("sockutil passed tests\n");

  socket_lib_deinit();
  return 0;
}

USUAL_MAIN

#endif // TEST_SOCKUTIL
